from colorsys import hls_to_rgb as _hls_to_rgb
from colorsys import hsv_to_rgb as _hsv_to_rgb
from colorsys import rgb_to_hls as _rgb_to_hls
from colorsys import rgb_to_hsv as _rgb_to_hsv
from sys import maxsize
from typing import Iterable, List, Tuple

from colorpedia.hexcodes import HEX_CODE_TO_NAMES, NAME_TO_HEX_CODE
from colorpedia.palettes import PALETTES


def cmyk_to_rgb(c: float, m: float, y: float, k: float) -> Tuple[int, int, int]:
    """Convert CMYK (Cyan Magenta Yellow Black) to RGB (Red Green Blue).

    :param c: Cyan (0.0 to 1.0 inclusive).
    :param m: Magenta (0.0 to 1.0 inclusive).
    :param y: Yellow (0.0 to 1.0 inclusive).
    :param k: Black/Key (0.0 to 1.0 inclusive).
    :return: RGB tuple.
    """
    r = 255 * (1.0 - c) * (1.0 - k)
    g = 255 * (1.0 - m) * (1.0 - k)
    b = 255 * (1.0 - y) * (1.0 - k)

    return round(r), round(g), round(b)


def rgb_to_cmyk(r: int, g: int, b: int) -> Tuple[float, float, float, float]:
    """Convert RGB (Red Green Blue) to CMYK (Cyan Magenta Yellow Black).

    :param r: Red (0 to 255 inclusive).
    :param g: Green (0 to 255 inclusive).
    :param b: Blue (0 to 255 inclusive).
    :return: CMYK tuple.
    """
    if r == 0 and g == 0 and b == 0:
        return 0, 0, 0, 1

    c = 1 - r / 255
    m = 1 - g / 255
    y = 1 - b / 255

    cmy_min = min(c, m, y)
    c = (c - cmy_min) / (1 - cmy_min)
    m = (m - cmy_min) / (1 - cmy_min)
    y = (y - cmy_min) / (1 - cmy_min)
    k = cmy_min

    return c, m, y, k


def hex_to_rgb(hex_code: str) -> Tuple[int, int, int]:
    """Convert hexadecimal color code to RGB (Red Green Blue).

    :param hex_code: Hex color code without the hash (#) prefix.
    :return: RGB tuple.
    """
    r = int(hex_code[:2], 16)
    g = int(hex_code[2:4], 16)
    b = int(hex_code[4:6], 16)
    return r, g, b


def rgb_to_hex(r: int, g: int, b: int) -> str:
    """Convert RGB (Red Green Blue) to hexadecimal color code.

    :param r: Red (0 to 255 inclusive).
    :param g: Green (0 to 255 inclusive).
    :param b: Blue (0 to 255 inclusive).
    :return: Hexadecimal color code.
    """
    return f"{r:02x}{g:02x}{b:02x}".upper()


def hsl_to_rgb(h: float, s: float, l: float) -> Tuple[int, int, int]:
    """Convert HSL (Hue Saturation Lightness) to RGB (Red Green Blue).

    :param h: Hue (0.0 to 1.0 inclusive).
    :param s: Saturation (0.0 to 1.0 inclusive).
    :param l: Lightness (0.0 to 1.0 inclusive).
    :return: RGB tuple.
    """
    r, g, b = _hls_to_rgb(h, l, s)
    return round(r * 255), round(g * 255), round(b * 255)


def rgb_to_hsl(r: int, g: int, b: int) -> Tuple[float, float, float]:
    """Convert RGB (Red Green Blue) to HSL (Hue Saturation Lightness).

    :param r: Red (0 to 255 inclusive).
    :param g: Green (0 to 255 inclusive).
    :param b: Blue (0 to 255 inclusive).
    :return: HSL tuple in degree angle and percent (e.g. 360.0, 95.0, 5.0).
    """
    h, l, s = _rgb_to_hls(r / 255, g / 255, b / 255)
    return h, s, l


def hsv_to_rgb(h: float, s: float, v: float) -> Tuple[int, int, int]:
    """Convert HSV (Hue Saturation Brightness) to RGB (Red Green Blue).

    :param h: Hue (0.0 to 1.0 inclusive).
    :param s: Saturation (0.0 to 1.0 inclusive).
    :param v: Brightness (0.0 to 1.0 inclusive).
    :return: RGB tuple.
    """
    r, g, b = _hsv_to_rgb(h, s, v)
    return round(r * 255), round(g * 255), round(b * 255)


def rgb_to_hsv(r: int, g: int, b: int) -> Tuple[float, float, float]:
    """Convert RGB (Red Green Blue) to HSV (Hue Saturation Brightness).

    :param r: Red (0 to 255 inclusive).
    :param g: Green (0 to 255 inclusive).
    :param b: Blue (0 to 255 inclusive).
    :return: HSV tuple.
    """
    return _rgb_to_hsv(r / 255, g / 255, b / 255)


def name_to_rgb(name: str) -> Tuple[int, int, int]:
    """Convert CSS3 color name to RGB (Red Green Blue).

    :param name: CSS3 color name.
    :return: RGB tuple.
    """
    try:
        return hex_to_rgb(NAME_TO_HEX_CODE[name])
    except KeyError:
        raise ValueError("Unknown color name (expecting a CSS3 color name)")


def rgb_to_names(r: int, g: int, b: int) -> Tuple[Tuple[str, ...], bool]:
    """Convert RGB (Red Green Blue) to the nearest CSS3 name(s).

    :param r: Red (0 to 255 inclusive).
    :param g: Green (0 to 255 inclusive).
    :param b: Blue (0 to 255 inclusive).
    :return: Color name(s) and a boolean indicating exact match.
    """
    try:
        return HEX_CODE_TO_NAMES[f"{r:02x}{g:02x}{b:02x}".upper()], True
    except KeyError:
        minimum_diff = maxsize
        nearest_names: Tuple[str, ...] = tuple()

        for hex_code, names in HEX_CODE_TO_NAMES.items():
            _r, _g, _b = hex_to_rgb(hex_code)

            dr = (_r - r) ** 2
            dg = (_g - g) ** 2
            db = (_b - b) ** 2
            diff = dr + dg + db

            if diff < minimum_diff:
                minimum_diff = diff
                nearest_names = names

        return nearest_names, False


def palette_to_rgbs(palette: str) -> List[Tuple[int, int, int]]:
    """Return the hex codes in the given palette in sorted order.

    :param palette: Palette name.
    :return: List of RGB tuples.
    """
    try:
        return [hex_to_rgb(hex_code) for hex_code in PALETTES[palette]]
    except KeyError:
        raise ValueError(f'Unknown color palette "{palette}"')


def hsl_to_rgb_shades(
    h: float, s: float, l: float, size: int
) -> Iterable[Tuple[int, int, int]]:
    """Yield RGBs ranging in lightness (shades) given an HSL.

    :param h: Hue (0.0 to 1.0 inclusive).
    :param s: Saturation (0.0 to 1.0 inclusive).
    :param l: Lightness (0.0 to 1.0 inclusive).
    :param size: Range (positive integer).
    :return: RGB tuple iterator.
    """
    if size > 1:
        for x in range(size):
            r, g, b = _hls_to_rgb(h, x / (size - 1), s)
            yield round(r * 255), round(g * 255), round(b * 255)
    else:
        r, g, b = _hls_to_rgb(h, l, s)
        yield round(r * 255), round(g * 255), round(b * 255)



import os
import shlex
import shutil
import subprocess
from dataclasses import dataclass
from json import dump as json_dump
from json import load as json_load
from pathlib import Path
from typing import Any, Dict, FrozenSet, Optional, Union

from colorpedia.exceptions import ConfigFileError, ConfigKeyError, ConfigValueError

CONFIG_DIR = Path.home() / ".config" / "colorpedia"
CONFIG_FILE = CONFIG_DIR / "config.json"
TMP_CONFIG_FILE = CONFIG_DIR / "config.json.tmp"

VIEW_KEYS = frozenset(("name", "rgb", "cmyk", "hex", "hsv", "hsl", "color"))
JSON_KEYS = frozenset(("is_name_exact", "name", "rgb", "cmyk", "hex", "hsv", "hsl"))
DEFAULT_SHADES_COUNT = 15
GET_VIEW_COLOR_HEIGHT = 10
GET_VIEW_COLOR_WIDTH = 20
LIST_VIEW_COLOR_WIDTH = 20


@dataclass
class Config:
    always_output_json: bool = False
    approx_name_suffix: str = "~"
    default_shades_count: int = DEFAULT_SHADES_COUNT
    display_degree_symbol: bool = False
    display_percent_symbol: bool = False
    get_view_color_height: int = GET_VIEW_COLOR_HEIGHT
    get_view_color_width: int = GET_VIEW_COLOR_WIDTH
    get_view_keys: FrozenSet[str] = VIEW_KEYS
    list_view_color_width: int = LIST_VIEW_COLOR_WIDTH
    list_view_keys: FrozenSet[str] = VIEW_KEYS
    json_keys: FrozenSet[str] = JSON_KEYS
    uppercase_hex_codes: bool = True

    def update(self, data: Dict[str, Any]) -> None:
        if type(data) != dict:
            raise ConfigFileError("Bad JSON: expecting an object")

        def validate_string(name: str) -> None:
            if type(getattr(self, name)) != str:
                raise ConfigValueError(name, "a string")

        def validate_boolean(name: str) -> None:
            if type(getattr(self, name)) != bool:
                raise ConfigValueError(name, "true or false")

        def validate_number(name: str) -> None:
            value = getattr(self, name)
            if not (type(value) == int and 1 <= value <= 100):
                raise ConfigValueError(name, "an integer between 1 and 100")

        def validate_view_keys(name: str) -> None:
            keys = getattr(self, name)
            if not (
                type(keys) in (list, frozenset)
                and len(keys) > 0
                and set(keys).issubset(VIEW_KEYS)
            ):
                raise ConfigValueError(
                    name, f"non-empty array of strings in {list(VIEW_KEYS)}"
                )

        def validate_json_keys(name: str) -> None:
            keys = getattr(self, name)
            if not (
                type(keys) in (list, frozenset)
                and len(keys) > 0
                and set(keys).issubset(JSON_KEYS)
            ):
                raise ConfigValueError(
                    name, f"non-empty array of strings in {list(JSON_KEYS)}"
                )

        for key, val in data.items():
            if hasattr(self, key):
                setattr(self, key, val)
            else:
                raise ConfigKeyError(key)

        validate_string("approx_name_suffix")
        validate_boolean("always_output_json")
        validate_boolean("display_degree_symbol")
        validate_boolean("display_percent_symbol")
        validate_boolean("uppercase_hex_codes")
        validate_number("default_shades_count")
        validate_number("get_view_color_height")
        validate_number("get_view_color_width")
        validate_number("list_view_color_width")
        validate_view_keys("get_view_keys")
        validate_view_keys("list_view_keys")
        validate_json_keys("json_keys")

        self.get_view_keys = frozenset(self.get_view_keys)
        self.list_view_keys = frozenset(self.list_view_keys)
        self.json_keys = frozenset(self.json_keys)

    def dump(self) -> Dict[str, Any]:
        data_dump = {}
        for key, value in self.__dict__.items():
            if isinstance(value, frozenset):
                data_dump[key] = list(value)
            else:
                data_dump[key] = value
        return data_dump

    def set_flags(
        self,
        json: Optional[bool] = None,
        all: Optional[bool] = None,
        units: Optional[bool] = None,
        shades: Optional[Union[bool, int]] = None,
    ) -> None:
        if json is True:
            self.always_output_json = True
        elif json is False:
            self.always_output_json = False

        if all is True:
            self.get_view_keys = VIEW_KEYS
            self.list_view_keys = VIEW_KEYS
            self.json_keys = JSON_KEYS

        if units is True:
            self.display_degree_symbol = True
            self.display_percent_symbol = True
        elif units is False:
            self.display_degree_symbol = False
            self.display_percent_symbol = False

        if shades is False:
            self.default_shades_count = 0
        elif shades is not True and shades is not None:
            self.default_shades_count = shades


def load_config_file() -> Config:  # pragma: no cover
    config = Config()
    try:
        with open(CONFIG_FILE, "r") as fp:
            data = json_load(fp)

    except FileNotFoundError:
        return config
    except ValueError as err:
        raise ConfigFileError("Bad JSON", err)
    except Exception as err:
        raise ConfigFileError(f"Cannot load {CONFIG_FILE}", err)
    else:
        config.update(data)
        return config


def save_config_file(config: Config) -> None:  # pragma: no cover
    data = config.dump()
    try:
        with open(CONFIG_FILE, "w") as fp:
            json_dump(data, fp, sort_keys=True, indent=2)

    except ValueError as err:
        raise ConfigFileError("Bad JSON", err)
    except Exception as err:
        raise ConfigFileError(f"Cannot save {CONFIG_FILE}", err)


def init_config_file() -> Config:  # pragma: no cover
    try:
        CONFIG_DIR.mkdir(parents=True, exist_ok=True)
    except Exception as err:
        raise ConfigFileError(f"Cannot create {CONFIG_DIR}", err)
    else:
        config = Config()
        save_config_file(config)
        return config


def edit_config_file(editor: Optional[str] = None) -> Config:  # pragma: no cover
    editor = editor or os.environ.get("VISUAL") or os.environ.get("EDITOR")
    if editor:
        editor = shlex.split(editor)[0]  # Prevent arbitrary code execution
    elif os.name == "nt":
        editor = "notepad"
    else:
        editor = "vi"
    try:
        shutil.copy(CONFIG_FILE, TMP_CONFIG_FILE)
        subprocess.check_call([editor, str(TMP_CONFIG_FILE)])

        with open(TMP_CONFIG_FILE, "r") as fp:
            raw_config = json_load(fp)

    except Exception as err:
        raise ConfigFileError(f"Cannot update {CONFIG_FILE}", err)
    else:
        config = Config()
        config.update(raw_config)
        save_config_file(config)
        return config
    finally:
        try:
            TMP_CONFIG_FILE.unlink()
        except FileNotFoundError:
            pass



from colorpedia.color import Color
from colorpedia.config import Config


def format_degree(value: float) -> str:
    return f"{value:<3.0f}"


def format_degree_with_unit(value: float) -> str:
    string = f"{value:.0f}°"
    return f"{string:<4s}"


def format_percent(value: float) -> str:
    return f"{value:<3.0f}"


def format_percent_with_unit(value: float) -> str:
    string = f"{value:.0f}%"
    return f"{string:<4s}"


def format_get_color(config: Config, r: int, g: int, b: int) -> str:
    h = config.get_view_color_height
    w = config.get_view_color_width
    line = f'\033[48;2;{r};{g};{b}m{" " * w}\033[0m'
    return "\n".join(line for _ in range(h))


def format_list_color(config: Config, r: int, g: int, b: int) -> str:
    return f'\033[48;2;{r};{g};{b}m{" " * config.list_view_color_width}\033[0m'


def format_name(config: Config, name: str, is_exact: bool) -> str:
    return name if is_exact else f"{name}{config.approx_name_suffix}"


def format_cmyk(config: Config, c: float, m: float, y: float, k: float) -> str:
    if config.display_percent_symbol:
        p = format_percent_with_unit
    else:
        p = format_percent

    return f"C:{p(c * 100)} M:{p(m * 100)} Y:{p(y * 100)} K:{p(k * 100)}"


def format_hsv(config: Config, h: float, s: float, v: float) -> str:
    if config.display_degree_symbol:
        d = format_degree_with_unit
    else:
        d = format_degree

    if config.display_percent_symbol:
        p = format_percent_with_unit
    else:
        p = format_percent

    return f"H:{d(h * 360)} S:{p(s * 100)} V:{p(v * 100)}"


def format_hsl(config: Config, h: float, s: float, l: float) -> str:
    if config.display_degree_symbol:
        d = format_degree_with_unit
    else:
        d = format_degree

    if config.display_percent_symbol:
        p = format_percent_with_unit
    else:
        p = format_percent

    return f"H:{d(h * 360)} S:{p(s * 100)} L:{p(l * 100)}"


def format_hex(config: Config, hex_code: str) -> str:
    if config.uppercase_hex_codes:
        return "#" + hex_code.upper()
    else:
        return "#" + hex_code.lower()


def format_rgb(config: Config, r: int, g: int, b: int) -> str:
    r_str = f"{r:<4d}" if config.display_degree_symbol else f"{r:<3d}"
    g_str = f"{g:<4d}" if config.display_percent_symbol else f"{g:<3d}"
    b_str = f"{b:<4d}" if config.display_percent_symbol else f"{b:<3d}"
    return f"R:{r_str} G:{g_str} B:{b_str}"


def format_get_view(config: Config, color: Color) -> str:
    keys = config.get_view_keys
    buf = []
    if "name" in keys:
        buf.append(f"Name : {format_name(config, color.name, color.is_name_exact)}")
    if "hex" in keys:
        buf.append(f"Hex  : {format_hex(config, color.hex)}")
    if "rgb" in keys:
        buf.append(f"RGB  : {format_rgb(config, *color.rgb)}")
    if "hsl" in keys:
        buf.append(f"HSL  : {format_hsl(config, *color.hsl)}")
    if "hsv" in keys:
        buf.append(f"HSV  : {format_hsv(config, *color.hsv)}")
    if "cmyk" in keys:
        buf.append(f"CMYK : {format_cmyk(config, *color.cmyk)}")
    if "color" in keys:
        buf.append("\n" + format_get_color(config, *color.rgb))
    return "\n".join(buf)


def format_list_view(config: Config, color: Color) -> str:
    keys = config.list_view_keys
    buf = []
    if "color" in keys:
        buf.append(format_list_color(config, *color.rgb))
    if "hex" in keys:
        buf.append(format_hex(config, color.hex))
    if "rgb" in keys:
        buf.append(format_rgb(config, *color.rgb))
    if "hsl" in keys:
        buf.append(format_hsl(config, *color.hsl))
    if "hsv" in keys:
        buf.append(format_hsv(config, *color.hsv))
    if "cmyk" in keys:
        buf.append(format_cmyk(config, *color.cmyk))
    if "name" in keys:
        buf.append(format_name(config, color.name, color.is_name_exact))
    return "|".join(buf)



from dataclasses import dataclass
from typing import Any, Dict, FrozenSet, Iterable, Set, Union

from colorpedia.converters import (
    hsl_to_rgb_shades,
    rgb_to_cmyk,
    rgb_to_hex,
    rgb_to_hsl,
    rgb_to_hsv,
    rgb_to_names,
)


@dataclass(eq=True, unsafe_hash=True)
class Color:
    r: int
    g: int
    b: int

    def __post_init__(self) -> None:
        self.rgb = (self.r, self.g, self.b)
        self.names, self.is_name_exact = rgb_to_names(*self.rgb)
        self.name = "/".join(self.names)
        self.hex = rgb_to_hex(*self.rgb)
        self.hsv = rgb_to_hsv(*self.rgb)
        self.hsl = rgb_to_hsl(*self.rgb)
        self.cmyk = rgb_to_cmyk(*self.rgb)

    def get_shades(self, size: int) -> Iterable["Color"]:
        h, s, l = self.hsl
        return (Color(*rgb) for rgb in hsl_to_rgb_shades(h, s, l, size))

    def get_dict(self, keys: Union[FrozenSet[str], Set[str]]) -> Dict[str, Any]:
        result: Dict[str, Any] = {}
        if "hex" in keys:
            result["hex"] = self.hex
        if "rgb" in keys:
            result["rgb"] = self.rgb
        if "hsl" in keys:
            result["hsl"] = self.hsl
        if "hsv" in keys:
            result["hsv"] = self.hsv
        if "cmyk" in keys:
            result["cmyk"] = self.cmyk
        if "name" in keys:
            result["name"] = self.name
        if "is_name_exact" in keys:
            result["is_name_exact"] = self.is_name_exact
        return result



import os
import sys
from distutils.util import strtobool
from json import dumps as json_dumps
from typing import Any, Callable, Dict, Iterable, Optional

from fire import Fire
from pkg_resources import get_distribution

from colorpedia.color import Color
from colorpedia.config import (
    CONFIG_FILE,
    Config,
    edit_config_file,
    init_config_file,
    load_config_file,
)
from colorpedia.converters import (
    cmyk_to_rgb,
    hex_to_rgb,
    hsl_to_rgb,
    hsv_to_rgb,
    name_to_rgb,
    palette_to_rgbs,
)
from colorpedia.exceptions import ColorpediaError
from colorpedia.formatters import format_get_view, format_list_view
from colorpedia.hexcodes import NAME_TO_HEX_CODE
from colorpedia.inputs import (
    normalize_degree_angle,
    normalize_hex_code,
    normalize_percent_value,
    validate_boolean_flag,
    validate_editor,
    validate_indent_width,
    validate_rgb_value,
    validate_shades_count,
)
from colorpedia.palettes import PALETTES


def prompt_user(question: str) -> bool:
    while True:
        sys.stdout.write(f"{question} [y/n] ")
        try:
            return strtobool(input().lower())
        except ValueError:
            print('Please respond with "y" or "n"\n')


def print_colors(config: Config, colors: Iterable[Color]) -> None:
    if config.always_output_json:
        print(json_dumps([c.get_dict(config.json_keys) for c in colors]))
    else:
        for color in colors:
            print(format_list_view(config, color))


def print_color(config: Config, color: Color) -> None:
    if config.default_shades_count:
        print_colors(config, color.get_shades(config.default_shades_count))

    elif config.always_output_json:
        print(json_dumps(color.get_dict(config.json_keys)))
    else:
        print(format_get_view(config, color))


def print_config(config: Config, sort: bool = True, indent: int = 2) -> None:
    print(json_dumps(config.dump(), sort_keys=sort, indent=indent))


def get_version(json: Optional[bool] = None) -> None:
    """Display Colorpedia CLI version.

    Colorpedia follows Semantic Versioning 2.0.0 (semver.org).

    :param json: Display in JSON format.
    """
    json = validate_boolean_flag(json)
    version = get_distribution("colorpedia").version
    print({"version": version} if json else version)


def init_config(force: bool = False) -> None:
    """Initialize or reset Colorpedia configuration.

    Configuration file is located at ~/.config/colorpedia/config.json.

    :param force: Overwrite existing configuration without asking user
        (default: False).
    """
    if CONFIG_FILE.exists() and not force:
        overwrite = prompt_user("Overwrite existing configuration?")
        if not overwrite:
            return
    print_config(init_config_file())


def show_config(sort: bool = True, indent: int = 2) -> None:
    """Display Colorpedia configuration.

    Configuration file is located at ~/.config/colorpedia/config.json.

    :param sort: Sort JSON keys (default: True).
    :param indent: JSON indent width (default: 2).
    """
    if not CONFIG_FILE.exists():
        print('Configuration not initialized. Run "color config init".')
    else:
        config = load_config_file()
        validate_boolean_flag(sort)
        validate_indent_width(indent)
        print_config(config, sort=sort, indent=indent)


def edit_config(editor: Optional[str] = None) -> None:
    """Edit Colorpedia configuration using a text editor.

    If --editor flag is not specified, $VISUAL and $EDITOR environment
    variables are checked. If they are not defined, notepad is used on
    Windows and vi on Linux/MacOS.

    :param editor: Text editor to use. Must be a shell executable command.
        Any extra command options are ignored.
    """
    if not CONFIG_FILE.exists():
        print('Configuration not initialized. Run "color config init".')
    else:
        validate_editor(editor)
        config = edit_config_file(editor)
        print_config(config)


def get_palette_func(name: str) -> Callable[..., None]:
    def function(
        json: Optional[bool] = None,
        all: bool = False,
        units: Optional[bool] = None,
    ) -> None:
        config = load_config_file()
        config.set_flags(
            json=validate_boolean_flag(json),
            all=validate_boolean_flag(all),
            units=validate_boolean_flag(units),
        )
        print_colors(config, [Color(*rgb) for rgb in palette_to_rgbs(name)])

    function.__doc__ = "\n".join(
        (
            f'Display colors in palette "{name}".',
            ":param json: Display in JSON format.",
            ":param all: Bypass user configuration and display all keys.",
            ":param units: Bypass user configuration and display units.",
        )
    )
    return function


def get_color_by_name_func(name: str, hex_code: str) -> Callable[..., None]:
    def function(
        shades: int = 0,
        json: Optional[bool] = None,
        all: bool = False,
        units: Optional[bool] = None,
    ) -> None:
        config = load_config_file()
        config.set_flags(
            shades=validate_shades_count(shades),
            json=validate_boolean_flag(json),
            all=validate_boolean_flag(all),
            units=validate_boolean_flag(units),
        )
        print_color(config, Color(*name_to_rgb(name)))

    function.__doc__ = "\n".join(
        (
            f"Display color {name} (hex code #{hex_code.upper()}).",
            ":param json: Display in JSON format.",
            ":param shades: Display different shades of the specified color.",
            ":param all: Bypass user configuration and display all keys.",
            ":param units: Bypass user configuration and display units.",
        )
    )
    return function


def get_color_by_cmyk(
    c: float,
    m: float,
    y: float,
    k: float,
    shades: int = 0,
    json: Optional[bool] = None,
    all: bool = False,
    units: Optional[bool] = None,
) -> None:
    """Look up colors by CMYK (Cyan Magenta Yellow Black) values.

    CMYK is a subtractive color model used in color printing. It refers
    to the four ink plates used in color printing: cyan, magenta, yellow,
    and key (black). Values must be between 0.0 and 100.0 inclusive.

    Usage examples:

        color cmyk 100 100 100 100
        color cmyk 0 0 0 0
        color cmyk 10 20 30 40 --shades --json --all --units

    :param c: Cyan % (0.0 to 100.0 inclusive).
    :param m: Magenta % (0.0 to 100.0 inclusive).
    :param y: Yellow % (0.0 to 100.0 inclusive).
    :param k: Black/Key % (0.0 to 100.0 inclusive).
    :param shades: Display different shades of the specified color.
    :param json: Display in JSON format.
    :param all: Bypass user configuration and display all keys.
    :param units: Bypass user configuration and display units.
    """
    config = load_config_file()
    config.set_flags(
        shades=validate_shades_count(shades),
        json=validate_boolean_flag(json),
        all=validate_boolean_flag(all),
        units=validate_boolean_flag(units),
    )
    c = normalize_percent_value(c)
    m = normalize_percent_value(m)
    y = normalize_percent_value(y)
    k = normalize_percent_value(k)
    print_color(config, Color(*cmyk_to_rgb(c, m, y, k)))


def get_color_by_hex(
    hex_code: str,
    shades: int = 0,
    json: Optional[bool] = None,
    all: bool = False,
    units: Optional[bool] = None,
) -> None:
    """Look up colors by hexadecimal (web) code.

    The hexadecimal code must be specified without the hash (#) prefix
    and match the regular expression: ^(?:[0-9a-fA-F]{3}){1,2}$

    Usage examples:

        color hex FFFFFF
        color hex FFF
        color hex FFFFFF --shades --json --all --units

    :param hex_code: Hex color code without the hash (#) prefix.
    :param shades: Display different shades of the specified color.
    :param json: Display in JSON format.
    :param all: Bypass user configuration and display all keys.
    :param units: Bypass user configuration and display units.
    """
    config = load_config_file()
    config.set_flags(
        shades=validate_shades_count(shades),
        json=validate_boolean_flag(json),
        all=validate_boolean_flag(all),
        units=validate_boolean_flag(units),
    )
    hex_code = normalize_hex_code(hex_code)
    print_color(config, Color(*hex_to_rgb(hex_code)))


def get_color_by_hsl(
    h: float,
    s: float,
    l: float,
    shades: int = 0,
    json: Optional[bool] = None,
    all: bool = False,
    units: Optional[bool] = None,
) -> None:
    """Look up colors by HSL (Hue Saturation Lightness) values.

    Hue is a point on the color wheel from 0 to 360 where 0 is red,
    120 is green, and 240 is blue. Saturation is a percentage value
    where 0% means faded, and 100% means full color. Lightness is a
    percentage value where 0% is black and 100% is white.

    Usage examples:

        color hsl 360 100 100
        color hsl 0 0 0
        color hsl 10 20 30 --shades --json --all --units

    :param h: Hue in degree angle (0.0 to 360.0 inclusive).
    :param s: Saturation % (0.0 to 100.0 inclusive).
    :param l: Lightness % (0.0 to 100.0 inclusive).
    :param shades: Display different shades of the specified color.
    :param json: Display in JSON format.
    :param all: Bypass user configuration and display all keys.
    :param units: Bypass user configuration and display units.
    """
    config = load_config_file()
    config.set_flags(
        shades=validate_shades_count(shades),
        json=validate_boolean_flag(json),
        all=validate_boolean_flag(all),
        units=validate_boolean_flag(units),
    )
    h = normalize_degree_angle(h)
    s = normalize_percent_value(s)
    l = normalize_percent_value(l)
    print_color(config, Color(*hsl_to_rgb(h, s, l)))


def get_color_by_hsv(
    h: float,
    s: float,
    v: float,
    shades: int = 0,
    json: Optional[bool] = None,
    all: bool = False,
    units: Optional[bool] = None,
) -> None:
    """Look up colors by HSV (Hue Saturation Brightness/Value) values.

    Hue is a point on the color wheel from 0 to 360 where 0 is red,
    120 is green, and 240 is blue. Saturation is a percentage value
    where 0% means faded, and 100% means full color. Brightness (or
    value) is a percentage value where 0% is black and 100% reveals
    the most color.

    Usage examples:

        color hsv 360 100 100
        color hsv 0 0 0
        color hsv 10 20 30 --shades --json --all --units

    :param h: Hue in degree angle (0.0 to 360.0 inclusive).
    :param s: Saturation % (0.0 to 100.0 inclusive).
    :param v: Brightness/Value % (0.0 to 100.0 inclusive).
    :param shades: Display different shades of the specified color.
    :param json: Display in JSON format.
    :param all: Bypass user configuration and display all keys.
    :param units: Bypass user configuration and display units.
    """
    config = load_config_file()
    config.set_flags(
        shades=validate_shades_count(shades),
        json=validate_boolean_flag(json),
        all=validate_boolean_flag(all),
        units=validate_boolean_flag(units),
    )
    h = normalize_degree_angle(h)
    s = normalize_percent_value(s)
    v = normalize_percent_value(v)
    print_color(config, Color(*hsv_to_rgb(h, s, v)))


def get_color_by_rgb(
    r: int,
    g: int,
    b: int,
    shades: int = 0,
    json: Optional[bool] = None,
    all: bool = False,
    units: Optional[bool] = None,
) -> None:
    """Look up colors by RGB (Red Green Blue) values.

    RGB is an additive color model where red, green, and blue lights are
    superimposed together to yield shades of colors. The RGB values must
    be integers between 0 to 255.

    Usage examples:

        color rgb 255 255 255
        color rgb 0 0 0
        color rgb 10 20 30 --shades --json --all --units

    :param r: Red (0 to 255 inclusive).
    :param g: Green (0 to 255 inclusive).
    :param b: Blue (0 to 255 inclusive).
    :param shades: Display different shades of the specified color.
    :param json: Display in JSON format.
    :param all: Bypass user configuration and display all keys.
    :param units: Bypass user configuration and display units.
    """
    config = load_config_file()
    config.set_flags(
        shades=validate_shades_count(shades),
        json=validate_boolean_flag(json),
        all=validate_boolean_flag(all),
        units=validate_boolean_flag(units),
    )
    r = validate_rgb_value(r)
    g = validate_rgb_value(g)
    b = validate_rgb_value(b)
    print_color(config, Color(r, g, b))


class MainCommand(Dict[str, Any]):
    """Colorpedia CLI.

    Look up colors using various models:

        color name green
        color hex FFFFFF
        color rgb 255 255 255
        color hsl 360 100 100
        color hsv 360 100 100
        color cmyk 100 100 100 100

    Display different shades:

        color name green --shades
        color hex FFFFFF --shades=5

    Look up color palettes:

        color palette molokai

    Control output with global flags:

        color name red --json --all --units

    Manage user configuration:

        color config init
        color config show
        color config edit
    """


class NameSubCommand(Dict[str, Any]):
    """Look up colors by CSS3 name."""


class PaletteSubCommand(Dict[str, Any]):
    """Look up color palettes."""


class ConfigSubCommand(Dict[str, Any]):
    """Manage CLI configuration."""


def entry_point(name: str) -> None:
    # Workaround for python-fire's argument parsing
    args = sys.argv[1:]
    if args and args[0] == "hex":
        for i in range(1, len(args)):
            if not args[i].startswith("-"):
                args[i] = f'"{args[i]}"'
    try:
        # We need this to get colors working on windows.
        os.system("")
        Fire(
            name=name,
            command=args,
            component=MainCommand(
                {
                    "version": get_version,
                    "config": ConfigSubCommand(
                        {
                            "init": init_config,
                            "edit": edit_config,
                            "show": show_config,
                        }
                    ),
                    "name": NameSubCommand(
                        {
                            name: get_color_by_name_func(name, hex_code)
                            for name, hex_code in NAME_TO_HEX_CODE.items()
                        }
                    ),
                    "palette": PaletteSubCommand(
                        {name: get_palette_func(name) for name in PALETTES.keys()}
                    ),
                    "cmyk": get_color_by_cmyk,
                    "hex": get_color_by_hex,
                    "hsl": get_color_by_hsl,
                    "hsv": get_color_by_hsv,
                    "rgb": get_color_by_rgb,
                }
            ),
        )
    except KeyboardInterrupt:
        print()
    except ColorpediaError as err:
        sys.stderr.write(f"{err}\n")
        sys.exit(1)


def entry_point_color() -> None:
    entry_point("color")


def entry_point_colorpedia() -> None:
    entry_point("colorpedia")



from typing import Optional


class ColorpediaError(Exception):
    """Generic Colorpedia exception."""


class ConfigFileError(ColorpediaError):
    """Configuration file cannot be accessed, created or updated."""

    def __init__(self, message: str, err: Optional[Exception] = None):
        if isinstance(err, OSError):
            message = f"{message}: {err.strerror} (errno: {err.errno})"
        elif err:
            message = f"{message}: {err}"
        super().__init__(message)


class ConfigKeyError(ColorpediaError):
    """Configuration key is invalid."""

    def __init__(self, key: str):
        super().__init__(f'Bad configuration key "{key}"')


class ConfigValueError(ColorpediaError):
    """Configuration value is invalid."""

    def __init__(self, key: str, exp: str):
        super().__init__(f'Bad value for configuration key "{key}" (expecting {exp})')


class InputValueError(ColorpediaError):
    """Invalid input value from user."""

    def __init__(self, name: str, exp: str):
        super().__init__(f"Bad {name} (expecting {exp})")



from colorpedia.color import Color
from colorpedia.converters import (
    cmyk_to_rgb,
    hex_to_rgb,
    hsl_to_rgb,
    hsl_to_rgb_shades,
    hsv_to_rgb,
    name_to_rgb,
    palette_to_rgbs,
    rgb_to_cmyk,
    rgb_to_hex,
    rgb_to_hsl,
    rgb_to_hsv,
    rgb_to_names,
)



from setuptools import find_packages, setup

with open("docs/README.md") as fp:
    long_description = fp.read()

setup(
    name="colorpedia",
    description="CLI for Color Lookups",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Joohwan Oh",
    author_email="joohwan.oh@outlook.com",
    url="https://github.com/joowani/colorpedia",
    keywords=["cli", "color", "terminal"],
    packages=find_packages(exclude=["tests"]),
    include_package_data=True,
    python_requires=">=3.6",
    license="MIT",
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    install_requires=[
        "dataclasses>=0.6; python_version < '3.7'",
        "fire>=0.3.1",
        "setuptools>=42",
        "setuptools_scm[toml]>=3.4",
    ],
    extras_require={
        "dev": [
            "black",
            "flake8>=3.8.4",
            "isort>=5.0.0",
            "mkdocs-material",
            "mypy>=0.790",
            "pre-commit>=2.9.3",
            "pytest>=6.0.0",
            "pytest-cov>=2.0.0",
        ],
    },
    entry_points={
        "console_scripts": [
            "color = colorpedia.cli:entry_point_color",
            "colorpedia = colorpedia.cli:entry_point_colorpedia",
        ],
    },
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: Unix",
        "Programming Language :: Python :: 3",
        "Topic :: Documentation :: Sphinx",
        "Topic :: Utilities",
    ],
)



import re
from typing import Optional, Union

from colorpedia.exceptions import InputValueError
from colorpedia.hexcodes import HEX_REGEX


def validate_indent_width(value: int) -> int:
    if type(value) == int and 0 <= value <= 8:
        return value
    raise InputValueError("indent width", "an integer between 0 and 8")


def validate_boolean_flag(value: Optional[bool]) -> Optional[bool]:
    if value is True or value is False or value is None:
        return value
    raise InputValueError("boolean flag", "True, False or no value")


def validate_shades_count(value: Union[bool, int]) -> int:
    if (type(value) in (bool, int)) and 0 <= value <= 100:
        return value
    raise InputValueError("shades count", "an integer between 0 and 100")


def validate_editor(value: Optional[str]) -> Optional[str]:
    if value is None or (type(value) == str and len(value) > 0 and " " not in value):
        return value
    raise InputValueError("editor", "a shell-executable command without whitespaces")


def validate_rgb_value(value: int) -> int:
    if type(value) == int and 0 <= value <= 255:
        return value
    raise InputValueError("RGB value", "an integer between 0 and 255")


def normalize_degree_angle(value: Union[float, int]) -> float:
    if (type(value) in (float, int)) and 0 <= value <= 360:
        return value / 360
    raise InputValueError("degree angle", "a float between 0.0 and 360.0")


def normalize_hex_code(value: Union[int, str]) -> str:
    if isinstance(value, str) and re.search(HEX_REGEX, value):
        return value if len(value) == 6 else "".join(c * 2 for c in value)
    raise InputValueError("hex code", f"a string matching {HEX_REGEX}")


def normalize_percent_value(value: Union[float, int]) -> float:
    if (type(value) in (float, int)) and 0 <= value <= 100:
        return value / 100
    raise InputValueError("percent value", "a float between 0.0 and 100.0")
